require 'spec_helper'


def deep_copy_check(left, right)
  left.each_with_index do |item, idx|
    expect(right[idx]).not_to be(item)
  end
end

class Context
end

describe QueryInterface::Client::LazyQuery do
  subject {QueryInterface::Client::LazyQuery}
  let(:model) {double("Dummy Model")}
  let(:transformations) {[{transformation: :filter, parameter: {field: "hase", value: "wuschel"}}]}

  context "construction" do

    it "initializes itself with empty parameters and a supplied model" do
      query = subject.new(model)
      expect(query.transformations).to eq([])
      expect(query.model).to eq(model)
    end

    it "honors passed transformations params" do
      query = subject.new(model, transformations)
      expect(query.transformations).to eq(transformations)
    end

    it "does not alter the originally passed transformations" do
      query = subject.new(model, transformations)
      deep_copy_check(transformations, query.transformations)
    end
  end

  context "copy" do
    it "provides a copy method cloning transformations onto a new instance" do
      query = subject.new(model, transformations)
      query_copy = query.copy
      expect(query_copy.transformations).to eq(query.transformations)
      deep_copy_check(query_copy.transformations, query.transformations)
    end
  end

  context "filtering" do
    it "should create a new instance with updated transformations" do
      query = subject.new(model)
      query_copy = subject.new(model)
      expect(query).to receive(:copy).and_return(query_copy)
      query.filter(a: :b)
      expect(query_copy.transformations).to eq([{transformation: :filter, parameter: {field: :a, value: :b}}])
    end
  end

  context "instancing" do
    it "sets the instance parameter" do
      query = subject.new(model)
      expect(query).to receive(:copy).and_return(query)
      query.instance(5)
      expect(query.transformations).to eq([{transformation: :instance, parameter: 5}])
    end
  end

  context "context" do
    let(:query) {subject.new(model)}
    before do
      expect(query).to receive(:copy).and_return(query)
      query.context(:context)
    end
    it "should set the correct result model class" do
      expect(query.result_model).to eq(Context)
    end
    it "sets the context parameter" do
      expect(query.transformations).to eq([{transformation: :context, parameter: :context}])
    end
  end

  context "with" do
    it "should create a new instance including additional fields" do
      query = subject.new(model)
      query_copy = subject.new(model)
      expect(query).to receive(:copy).and_return(query_copy)
      query.with(:c)
      expect(query_copy.transformations).to eq([{transformation: :with, parameter: {field: :c}}])
    end
  end

  context "order" do
    it "should create a new instance includuing order fields" do
      query = subject.new(model)
      query_copy = subject.new(model)
      expect(query).to receive(:copy).and_return(query_copy)
      query.order("-something")
      expect(query_copy.transformations).to eq([{transformation: :order, parameter: "-something"}])
    end
  end

  context "chaining" do
    it "allows chaining transformations in order of appearance" do
      query = subject.new(model)
      query_copy = query.filter(a: :b, c: :d).filter(e: :f).with(:x, :y).instance(12).context(:context)
      expect(query_copy.transformations).to eq(
        [
          {transformation: :filter, parameter: {field: :a, value: :b}},
          {transformation: :filter, parameter: {field: :c, value: :d}},
          {transformation: :filter, parameter: {field: :e, value: :f}},
          {transformation: :with, parameter: {field: :x}},
          {transformation: :with, parameter: {field: :y}},
          {transformation: :instance, parameter: 12},
          {transformation: :context, parameter: :context}
        ]
      )
    end
  end

  context 'first' do
    it 'gets the first object' do
      query = subject.new(model)
      result = double('result')
      expect(query).to receive(:one).with(:first).and_return(result)
      expect(query.first).to be(result)
    end
  end

  context 'last' do
    it 'gets the last object' do
      query = subject.new(model)
      result = double('result')
      expect(query).to receive(:one).with(:last).and_return(result)
      expect(query.last).to be(result)
    end
  end

  context 'one' do
    it 'gets one object' do
      query = subject.new(model)
      result = double('result')
      expect(query).to receive(:copy).and_return(query)
      expect(query).to receive(:do_query).and_return(result)
      expect(query.result).to be_nil
      expect(query.send(:one, :anyone)).to be(result)
      expect(query.transformations).to eq([{transformation: :anyone, parameter: nil}])
    end
    it "uses the cached result" do
      query = subject.new(model)
      query.result = ["a", "b", "c"]
      expect(query).not_to receive(:do_query)
      expect(query.first).to eq("a")
    end
  end

  context "evaluate" do
    let(:transformations) {[]}
    context "without instance set" do
      before do
        expect(model).to receive(:get_raw)
        .with(path: :query, params: {transformations: JSON.dump(self.transformations)})
        .and_return({parsed_data: {data: ["result object"]}})
        allow(model).to receive(:new_collection).and_return(["result object"])
      end

      it "add an evaluation transformation" do
        query = subject.new(model)
        query.evaluate
        expect(query.transformations).to eq(transformations)
      end

      it "gets results via do_query and caches the result" do
        query = subject.new(model)
        expect(query.evaluate).to eq(["result object"])
        expect(query.result).to eq(["result object"])
      end

      it "doesn't query the api twice" do
        query = subject.new(model)
        result = query.evaluate
        result_second = query.evaluate
        expect(result).to be(result_second)
      end
    end
  end

  context "ids" do
    let(:transformations) { [{transformation: :map_ids, parameter: nil}] }

    it "adds a map_ids transformation" do
      query = subject.new(model)
      expect(query).to receive(:copy).and_return(query)
      allow(query).to receive(:do_raw_query).and_return({parsed_data: {data: [1,2,3]}})
      query.ids
      expect(query.transformations).to eq(self.transformations)
    end

    it "returns the data of the parsed query" do
      query = subject.new(model)
      expect(query).to receive(:copy).and_return(query)
      expect(query).to receive(:do_raw_query).and_return({parsed_data: {data: [1,2,3]}})
      expect(query.ids).to eq([1, 2, 3])
    end
  end

  context "count" do
    let(:transformations) {[{transformation: :count, parameter: nil}]}

    it "adds a count transformation" do
      query = subject.new(model)
      expect(query).to receive(:copy).and_return(query)
      expect(query).to receive(:do_raw_query).and_return({parsed_data: {data: {count: 42}}})
      expect(query.count).to eq(42)
      expect(query.transformations).to eq(self.transformations)
    end

    it "uses cached result for counting" do
      query = subject.new(model)
      query.result = ["a", "b", "c"]
      expect(query).not_to receive(:do_raw_query)
      expect(query.count).to eq(3)
    end
  end

  context "paginate" do
    let(:transformations) {[{transformation: :paginate, parameter: {page: 1, per_page: 10}}]}


    it "adds a paginate transformation" do
      query = subject.new(model)
      expect(query).to receive(:copy).and_return(query)
      objects = (1..10).to_a
      allow(model).to receive(:new_from_hash).and_return(*objects)
      expect(query).to receive(:do_raw_query).and_return({parsed_data: {data: {objects: objects, total: 15}, errors: []}})
      result = query.paginate
      expect(query.transformations).to eq(self.transformations)
      expect(result).to eq(objects)
      result.is_a?(RsPaginator::Collection)
      expect(result.total).to eq(15)
    end
  end

  context "method_missing" do
    it "delegates unknown methods to the result of evaluate" do
      query = subject.new(model)
      result = double("result")
      expect(query).to receive(:evaluate).and_return(result)
      expect(result).to receive(:frobnicate!).with("uargh!")
      query.frobnicate!("uargh!")
    end
  end

  context "do queries" do
    it "queries the api raw" do
      query = subject.new(model)
      expect(model).to receive(:get_raw).with(path: :query, params: {transformations: "[]"})
      query.do_raw_query
    end

    it "executes the actual query and creates a collection where appropriate" do
      query = subject.new(model)
      data = {parsed_data: {data: [1,2,3]}}
      expect(query).to receive(:do_raw_query).and_return(data)
      expect(query).to receive(:instantiate_collection).with(data[:parsed_data])
      query.do_query
    end

    it "executes the actual query and creates an object where appropriate" do
      query = subject.new(model)
      data = {parsed_data: {data: {id: 1, bunny: 'wuschel'}}}
      expect(query).to receive(:do_raw_query).and_return(data)
      expect(query).to receive(:instantiate).with(data[:parsed_data][:data])
      query.do_query
    end

  end

  context "to_json" do
    let(:transformations) {[]}
    let(:result) { double("result") }
    before do
      expect(model).to receive(:get_raw)
      .with(path: :query, params: {transformations: JSON.dump(self.transformations)})
      .and_return({parsed_data: {data: ["result object"]}})
      allow(model).to receive(:new_collection).and_return(result)
    end
    it "calls to_json on the evaluated result" do
      query = subject.new(model)
      expect(result).to receive(:to_json)
      query.to_json
    end
  end

  context "to_param" do
    let(:transformations) {[]}
    let(:result) { double("result") }
    before do
      expect(model).to receive(:get_raw)
      .with(path: :query, params: {transformations: JSON.dump(self.transformations)})
      .and_return({parsed_data: {data: ["result object"]}})
      allow(model).to receive(:new_collection).and_return(result)
    end
    it "calls to_key on the evaluated result if persisted" do
      query = subject.new(model)
      expect(result).to receive(:persisted?).and_return(true)
      expect(result).to receive(:to_key).and_return([5])
      expect(query.to_param).to eq("5")
    end

  end

end
